package com.agiledeveloper;

import java.lang.invoke.*;

public class Sample {
  public static void tick() {
    System.out.println("tick");
  }

  public static void tock() {
    System.out.println("tock");
  }

  static MethodHandle tickHandle;
  static MethodHandle tockHandle;

  public static void main(String[] args) throws Throwable {
    CallSite callSite = createCallSite();

    MethodHandle methodHandle = callSite.dynamicInvoker();

    for (int i = 0; i < 10; i++) {
      methodHandle.invoke();
      switchHandle(callSite, i);
      Thread.sleep(1000);
    }
  }

  private static void switchHandle(CallSite callSite, int index) throws NoSuchMethodException, IllegalAccessException {
    if (index % 2 == 0)
      callSite.setTarget(tockHandle);
    else
      callSite.setTarget(tickHandle);

    if (index == 8) {
      Runnable runnable = () -> System.out.println("tada");
      MethodHandle dynamicHandle =
          MethodHandles.lookup().findVirtual(
              runnable.getClass(), "run", MethodType.methodType(void.class)).bindTo(runnable);
      callSite.setTarget(dynamicHandle);
    }
  }

  private static CallSite createCallSite() throws NoSuchMethodException, IllegalAccessException {
    MethodHandles.Lookup lookup = MethodHandles.lookup();
    tickHandle = lookup.findStatic(Sample.class, "tick", MethodType.methodType(void.class));
    tockHandle = lookup.findStatic(Sample.class, "tock", MethodType.methodType(void.class));

    return new MutableCallSite(tickHandle);
  }
}
/*
tick
tock
tick
tock
tick
tock
tick
tock
tick
tada
*/
